/*         ______   ___    ___ 
 *        /\  _  \ /\_ \  /\_ \ 
 *        \ \ \L\ \\//\ \ \//\ \      __     __   _ __   ___ 
 *         \ \  __ \ \ \ \  \ \ \   /'__`\ /'_ `\/\`'__\/ __`\
 *          \ \ \/\ \ \_\ \_ \_\ \_/\  __//\ \L\ \ \ \//\ \L\ \
 *           \ \_\ \_\/\____\/\____\ \____\ \____ \ \_\\ \____/
 *            \/_/\/_/\/____/\/____/\/____/\/___L\ \/_/ \/___/
 *                                           /\____/
 *                                           \_/__/
 *
 *      Joystick driver for the Gravis GrIP.
 *      (Xterminator, GamepadPro, Stinger, ?)
 *
 *      By Robert J. Ragno, using the Gravis SDK and the other
 *      Allegro drivers as references.
 *
 */

/* TODO: 
     - Test on more devices
     - Add support for restricting joystick to 4-directions
     - Look into dropping the grip.gll support (with asm code)
     - Look into providing direction names, not just axis names
     - (analog detection)
*/

#ifndef DJGPP
#error This file should only be used by the djgpp version of Allegro
#endif

#include <stdlib.h>
#include <stdio.h>
#include <file.h>
#include <string.h>
#include <dir.h>

#include "allegro.h"
#include "grip.h"


/* Include test code, or not */
#define TEST_CODE     (1)

/* define the following if your code can't handle a complex
   joystick with multiple sticks. Though the Allegro stick
   limit becomes a big problem... Maybe an alternate driver
   is in order.
   *** not active in this code right now ***
*/
#define ONE_STICK_LIMIT     (0)

/* set the following to 0 if you want to skip unloading the
   grip.gll library. This shouldn't cause any real problems,
   especially if the program exits anyway, and it may prevent
   crashing due to undetected bugs in code outside this file.
*/
#define UNLOAD_GRIP         (0)

/* Define this to treat throttles as buttons... */
/* Not so needed, since they only trigger in one direction on
   their axis. If there is a need to map them to *both* analog
   controls and buttons, I'll think about it...
*/
#define THROTTLE_AS_BUTTON  (0) 

/* If this is true, throttles will be 'packed' to fit in the
smallest number of pseudo-sticks possible.
*/
#define PACK_THROTTLES      (1)

/* enable support for each type of control */
#define ENABLE_AXIS         (1)
#define ENABLE_ANALOG       (1)
#define ENABLE_POV          (1)
#define ENABLE_THROTTLE     (1)

/***** End of Compile-Time Parameters *****/

/* Handle if the limits aren't raised... */
#ifndef MAX_JOYSTICKS
/*#define MAX_JOYSTICKS      4*/
/* Problematic - can't be statically determined */
#define MAX_JOYSTICKS      ((int)sizeof(joy) / (int)sizeof(joy[0]))
#endif
#ifndef MAX_AXIS          
/*#define MAX_AXIS           3*/
#define MAX_AXIS           ((int)sizeof(joy[0].stick[0].axis) / (int)sizeof(joy[0].stick[0].axis[0]))
#endif
#ifndef MAX_BUTTONS       
/*#define MAX_BUTTONS       12*/
#define MAX_BUTTONS       ((int)sizeof(joy[0].button) / (int)sizeof(joy[0].button[0]))
#endif
#ifndef MAX_STICKS        
/*#define MAX_STICKS         4*/
#define MAX_STICKS         ((int)sizeof(joy[0].stick) / (int)sizeof(joy[0].stick[0]))
#endif


/* driver functions */
static int   grip_init(void);
static void  grip_exit(void); 
static int   grip_poll(void);
static int   grip_link(void);
static int   grip_unlink(void);
static int   read_grip(const int joy_num);

/*** Are the string issues documented anywhere? I can't find any 
specifications that relate to any of this. I suppose for now I will
just try to be safe and malloc (ick) any strings passed as well as
passing them through get_config_text (whatever that does).
***/

struct text_heap_s;
typedef struct text_heap_s  text_heap;
typedef text_heap *         text_heap_p;

struct text_heap_s
{
   char      *data;
   text_heap *last;
};

text_heap * texth = NULL;
static char * new_text(const char * const txt);
static void   free_text(void);
static int    text_heap_size(const text_heap * const th);

/* I have no idea what is right here. */
/*#define FIXUP_STR(foo)    get_config_text(new_text(foo))*/
/*#define FIXUP_STR(foo)    new_text(foo)*/
#define FIXUP_STR(foo)    get_config_text(foo)

/* hack to simplify axis names... */
#define SIMPLIFY_AXIS(foo)    if (foo[0] == 'X' && foo[1] == '_')  foo[1] = '\0'; \
            if (foo[0] == 'Y' && foo[1] == '_')  foo[1] = '\0';


#if TEST_CODE
/* testing functions */
static void    fprint_axis(FILE * const fp, const JOYSTICK_AXIS_INFO a);
static void    fprint_stick(FILE * const fp, const JOYSTICK_STICK_INFO s);
static void    fprint_button(FILE * const fp, const JOYSTICK_BUTTON_INFO button);
static void    fprint_joystick(FILE *const fp, const JOYSTICK_INFO j);
static void    fprint_joy(FILE *const fp);
static char *  binstr_long(unsigned long bitfield);
static char *  binstr(unsigned int bitfield);
static void    fprint_class(FILE * const fp, const GRIP_SLOT slot, const GRIP_CLASS gclass);
static void    fprint_slot(FILE * const fp, const GRIP_SLOT slot);
static void    fprint_grip(FILE * const fp);
#endif /*TEST_CODE*/


/*** Internal Variables for extra joystick data ***/
static GRIP_SLOT   slot_index[MAX_JOYSTICKS];
static GRIP_VALUE  max_value[MAX_JOYSTICKS][(int)GRIP_CLASS_MAX + 1];
static int         throttle_button_start[MAX_JOYSTICKS];
static GRIP_CLASS  stick_class[MAX_JOYSTICKS][MAX_STICKS];
static int         fourway[MAX_JOYSTICKS][MAX_STICKS];



JOYSTICK_DRIVER joystick_grip =
{
   JOY_TYPE_GRIP,
   "GrIP",      /* Perhaps fetch name too? But varies per device... */
   "GrIP",
   "GrIP",
   grip_init,
   grip_exit,
   grip_poll,
   NULL, NULL,  /* no saving/loading data */
   NULL, NULL   /* no calibrating */
};

/* Only allows 4-way movement on digital pad */
JOYSTICK_DRIVER joystick_grip4 =
{
   JOY_TYPE_GRIP4,
   "GrIP",      /* Perhaps fetch name too? But varies per device... */
   "GrIP",
   "GrIP 4-way",
   grip_init,
   grip_exit,
   grip_poll,
   NULL, NULL,  /* no saving/loading data */
   NULL, NULL   /* no calibrating */
};



/* int read_grip(int joy_num);
 *  Reads the GrIP controller pointed to by joy_num (0 or 1) and 
 *  sets its state; no useful return value currently.
 */
/* Does not refresh the GrIP state!!! Not a bug, just mentioning it. */
static int read_grip(const int j)
{
   int b;
   GRIP_SLOT slot_num;
   int state;
   GRIP_BITFIELD buttons_packed;
   int s;
   GRIP_CLASS last_class;
   GRIP_INDEX axis;
   int a;
   
   slot_num = slot_index[j];

   /* Read Sticks */
   /* sticks must be grouped by type for axis grouping to work! */
   axis = (GRIP_INDEX)0;
   last_class = stick_class[j][0];
   for (s = 0; s < joy[j].num_sticks; s++)
   {
      int reverse_factor; /* messy way to account for up/down reversal */
      if (joy_type == JOY_TYPE_GRIP4 &&
          stick_class[j][s] == GRIP_CLASS_AXIS && 
          joy[j].stick[s].num_axis == 2)
      {
         fourway[j][s] = joy[j].stick[s].axis[0].d1 + 
                         (joy[j].stick[s].axis[0].d2 << 1) +
                         (joy[j].stick[s].axis[1].d1 << 2) +
                         (joy[j].stick[s].axis[1].d2 << 3);
      }
   	  /* Reset axis count for new stick type */
   	  if (stick_class[j][s] != last_class)  axis = (GRIP_INDEX)0;
      last_class = stick_class[j][s];
      
      for (a = 0; a < joy[j].stick[s].num_axis; a++)
      {
         if ( a == 1 && (stick_class[j][s] != GRIP_CLASS_THROTTLE) )
            reverse_factor = -1;
         else
            reverse_factor = 1;
         state = (int)GrGetValue(slot_num, stick_class[j][s], axis);
         joy[j].stick[s].axis[a].pos = 
               state * 256 / max_value[j][(int)stick_class[j][s]] - 128;
         joy[j].stick[s].axis[a].pos *= reverse_factor;
         /* Don't trigger down direction if throttle */
         if (stick_class[j][s] != GRIP_CLASS_THROTTLE)
            joy[j].stick[s].axis[a].d1 = joy[j].stick[s].axis[a].pos < -64;
         joy[j].stick[s].axis[a].d2 = joy[j].stick[s].axis[a].pos > 64;
         axis++;
      }
   }

   /* Handle 4-way restriction */
   if (joy_type == JOY_TYPE_GRIP4)
   {
      for (s = 0; s < joy[j].num_sticks; s++)
      {
         if (stick_class[j][s] != GRIP_CLASS_AXIS || joy[j].stick[s].num_axis != 2)
            continue;
      
         if ((fourway[j][s] & 1) && joy[j].stick[s].axis[0].d1)
         {
            joy[j].stick[s].axis[0].d2 = 0;
            joy[j].stick[s].axis[1].d1 = 0;
            joy[j].stick[s].axis[1].d2 = 0;
         }
         if ((fourway[j][s] & 2) && joy[j].stick[s].axis[0].d2)
         {
            joy[j].stick[s].axis[0].d1 = 0;
            joy[j].stick[s].axis[1].d1 = 0;
            joy[j].stick[s].axis[1].d2 = 0;
         }
         if ((fourway[j][s] & 4) && joy[j].stick[s].axis[1].d1)
         {
            joy[j].stick[s].axis[1].d2 = 0;
            joy[j].stick[s].axis[0].d1 = 0;
            joy[j].stick[s].axis[0].d2 = 0;
         }
         if ((fourway[j][s] & 8) && joy[j].stick[s].axis[1].d2)
         {
            joy[j].stick[s].axis[1].d1 = 0;
            joy[j].stick[s].axis[0].d1 = 0;
            joy[j].stick[s].axis[0].d2 = 0;
         }
      } /* next s */
   } /* end if JOY_TYPE_GRIP4 */
   
   /* Read Buttons */
   buttons_packed = GrGetPackedValues((GRIP_SLOT)slot_num, GRIP_CLASS_BUTTON, 
	      (GRIP_INDEX)0, (GRIP_INDEX)(throttle_button_start[j] - 1));
   for (b=0; b < throttle_button_start[j]; b++) 
   {
      joy[j].button[b].b = !(!(buttons_packed & (0x01 << b)));
   }

   /* Read Throttle-Buttons */
   if (THROTTLE_AS_BUTTON)
   {
      for (b = throttle_button_start[j]; b < joy[j].num_buttons; b++)
      {
         if (3*(int)GrGetValue(slot_num, GRIP_CLASS_THROTTLE, 
                               (GRIP_INDEX)(b-throttle_button_start[j])) 
             > 2*(int)max_value[j][GRIP_CLASS_THROTTLE])
         {
         	joy[j].button[b].b = 1;
         } else
         {
         	joy[j].button[b].b = 0;
         }
            
      }
   }

   return 0;
}


/* grip_init:
 *  Initializes the driver.
 */
static int grip_init()
{
   char *name_x   = "X";
   char *name_y   = "Y";
   char *name_b[] = { "L", "R", "A", "B", "C", "D", "E", "F", "Select", "Start",
                      "B1", "B2", "B3", "B4", "B5", "B6", "B7", "B8", "B9", 
                      "B10", "B11", "B12", "B13", "B14", "B15", "B16", "B17", "B18" };
   char *name_pov_gen      = "Hat";
   char *name_stick_gen    = "Stick";
   char *name_analog_gen   = "Analog";
   char *name_throttle_gen = "Throttle";
   char tmpstr[127];
   int i = 0;
   GRIP_BITFIELD slots = 0;
   FILE *foo = NULL;

   /* The file grip.out is generated with debugging output if TEST_CODE
      is set to true at the top of this file.
   */
#if (TEST_CODE)
   foo = fopen("grip.out", "wt");
   if (foo && ferror(foo))
   {
      (void)fclose(foo);
      foo = NULL;
   }
   if (foo)  fprintf(foo, "Allegro GrIP debug data\n\n");
#endif /*TEST_CODE*/

   if (grip_link())
   {
      if (foo)  fprintf(foo, "Could not find GRIP.GLL\n");
      if (foo)  (void)fclose(foo);
      sprintf(allegro_error, "Could not find GRIP.GLL\n");
      return -1;
   }

   if (!GrInitialize())
   {
      if (foo)  fprintf(foo, "Could not initialize GrIP system.\n");
      if (foo)  (void)fclose(foo);
      sprintf(allegro_error, "Could not initialize GrIP system.\n");
      (void)grip_unlink();
      return -1;
   }

   (void)GrRefresh(0);  /* Probably unneeded, but still... */

   /* Assumes there are no joysticks but us. I don't like that, but
      the other drivers are doing it... And it might be the only way anyway.
   */
   num_joysticks = 0;
   slots = GrGetSlotMap();

   /* Fill slot_index with the slot number of each joystick found */
   for (i = 1; i <= 12; i++)
   {
     if (slots & (0x01 << i))
     {
     	slot_index[num_joysticks]  = (GRIP_SLOT)i;
     	num_joysticks++;
     	if (num_joysticks >= MAX_JOYSTICKS) break;
     }
   }
   if (num_joysticks == 0)
   {
      if (foo)  fprintf(foo, "No GrIP devices available.\n");
      if (foo)  (void)fclose(foo);
      sprintf(allegro_error, "No GrIP devices available.\n");
      (void)grip_exit();
      return -1;
   }

   /* Loop over each joystick (active GrIP slot) */
   for (i=0; i < num_joysticks; i++) {
   	  GRIP_BITFIELD class_map = 0;
   	  
      joy[i].flags = JOYFLAG_DIGITAL | JOYFLAG_SIGNED;
   	  class_map = GrGetClassMap(slot_index[i]);

      /* Standard GrIP features:                */
      /* AXIS ** POV HAT ** VELOCITY ** (ANALOG) ** (THROTTLE) ** BUTTONS */

      joy[i].num_sticks = 0;

      /* Check for an axis... (digital control) */
      if (ENABLE_AXIS && (class_map & (1uL << GRIP_CLASS_AXIS)))
      {
         int last_axis = 0;
         int stick = 0;

         last_axis = (int)GrGetMaxIndex(slot_index[i], GRIP_CLASS_AXIS);
         joy[i].num_sticks = last_axis/2 + 1;
         if (joy[i].num_sticks > MAX_STICKS)
         {
            joy[i].num_sticks = MAX_STICKS;
            last_axis = MAX_STICKS * 2;
         }
     	 /*axis_max[i]    = GrGetMaxValue(slot_index[i], GRIP_CLASS_AXIS);*/
         max_value[i][(int)GRIP_CLASS_AXIS] = 
               GrGetMaxValue(slot_index[i], GRIP_CLASS_AXIS);
         /* loop over each actual stick on the device (often just 1) */
         /* We arbitrarily split the axes up into pairs... GrIP gives no
            info on how this can be identified
          */
         for (stick = 0; (stick < joy[i].num_sticks); stick++)
         {
            stick_class[i][stick] = GRIP_CLASS_AXIS;

         	 /* name stick... :( */
            strcpy(tmpstr, name_stick_gen);
            if (joy[i].num_sticks > 1)
            {
               tmpstr[strlen(tmpstr)+2] = '\0';
               tmpstr[strlen(tmpstr)+1] = '1' + stick;
               tmpstr[strlen(tmpstr)] = ' ';
            }
            joy[i].stick[stick].name = FIXUP_STR(tmpstr);
            joy[i].stick[stick].flags = JOYFLAG_DIGITAL | JOYFLAG_SIGNED;
            /* name axes */
            joy[i].stick[stick].num_axis = ((stick * 2) == last_axis) ? 1 : 2;
            GrGetControlName(slot_index[i], GRIP_CLASS_AXIS,
                  (GRIP_INDEX)(2*stick), tmpstr);
            if (tmpstr[0] == '\0')  strcpy(tmpstr, name_x);
            SIMPLIFY_AXIS(tmpstr);
            joy[i].stick[stick].axis[0].name = FIXUP_STR(tmpstr);
            joy[i].stick[stick].axis[0].pos  = 0;
            joy[i].stick[stick].axis[0].d1  = 0;
            joy[i].stick[stick].axis[0].d2  = 0;

            if (joy[i].stick[stick].num_axis == 2)
            {
               GrGetControlName(slot_index[i], GRIP_CLASS_AXIS,
                     (GRIP_INDEX)(2*stick + 1), tmpstr);
               if (tmpstr[0] == '\0')  strcpy(tmpstr, name_y);
               SIMPLIFY_AXIS(tmpstr);
               joy[i].stick[stick].axis[1].name = FIXUP_STR(tmpstr);
               joy[i].stick[stick].axis[1].pos  = 0;
               joy[i].stick[stick].axis[1].d1  = 0;
               joy[i].stick[stick].axis[1].d2  = 0;
            }
         } /* next stick */
      } /* endif axes */


      /* Check for Analog pads (as in the Xterminator). 
         Assume they are in pairs of axes. 
       */
      if (ENABLE_ANALOG &&
          (class_map & (1uL << GRIP_CLASS_ANALOG)) && 
          ((int)GrGetMaxIndex(slot_index[i], GRIP_CLASS_ANALOG) > 0))
      {
         int num_pads = 0; 
         int stick = 0;
         int analog_start;

     	 /*analog_max[i]  = GrGetMaxValue(slot_index[i], GRIP_CLASS_ANALOG);*/
         max_value[i][(int)GRIP_CLASS_ANALOG] = 
               GrGetMaxValue(slot_index[i], GRIP_CLASS_ANALOG);
         num_pads    = ((int)GrGetMaxIndex(slot_index[i], GRIP_CLASS_ANALOG) + 1) / 2;
         analog_start = joy[i].num_sticks;
         joy[i].num_sticks += num_pads;
         if (joy[i].num_sticks > MAX_STICKS)  
         {
         	num_pads -= joy[i].num_sticks - MAX_STICKS;
         	joy[i].num_sticks = MAX_STICKS;
         }
         
         /* loop over each pair of analog pads on the device (often just 1) */
         for (stick = analog_start; stick < analog_start + num_pads; stick++)
         {
            stick_class[i][stick] = GRIP_CLASS_ANALOG;

         	 /* name pad... :( */
            strcpy(tmpstr, name_analog_gen);
            if (num_pads > 1)
            {
               tmpstr[strlen(tmpstr)+2] = '\0';
               tmpstr[strlen(tmpstr)+1] = '1' + stick - analog_start;
               tmpstr[strlen(tmpstr)] = ' ';
            }
            joy[i].stick[stick].name = FIXUP_STR(tmpstr);
            joy[i].stick[stick].flags = JOYFLAG_ANALOG | JOYFLAG_SIGNED;
            /* name axes */
            joy[i].stick[stick].num_axis = 2;
            GrGetControlName(slot_index[i], GRIP_CLASS_ANALOG,
                  (GRIP_INDEX)(2*(stick - analog_start)), tmpstr);
            if (tmpstr[0] == '\0')  strcpy(tmpstr, name_x);
            SIMPLIFY_AXIS(tmpstr);
            joy[i].stick[stick].axis[0].name = FIXUP_STR(tmpstr);
            joy[i].stick[stick].axis[0].pos  = 0;
            joy[i].stick[stick].axis[0].d1  = 0;
            joy[i].stick[stick].axis[0].d2  = 0;

            GrGetControlName(slot_index[i], GRIP_CLASS_ANALOG,
                  (GRIP_INDEX)(2*(stick - analog_start) + 1), tmpstr);
            if (tmpstr[0] == '\0')  strcpy(tmpstr, name_y);
            SIMPLIFY_AXIS(tmpstr);
            joy[i].stick[stick].axis[1].name = FIXUP_STR(tmpstr);
            joy[i].stick[stick].axis[1].pos  = 0;
            joy[i].stick[stick].axis[1].d1  = 0;
            joy[i].stick[stick].axis[1].d2  = 0;
         } /* next pad */
     } /* endif analogs */


      /* Check for a Point-of-View Hat...    */
      /*   Assume they are in pairs of axes. */
      if (ENABLE_POV &&
          (class_map & (1uL << GRIP_CLASS_POV_HAT)) && 
          ((int)GrGetMaxIndex(slot_index[i], GRIP_CLASS_POV_HAT) > 0))
      {
         int num_hats = 0; 
         int stick = 0;
         int pov_start;

     	 /*pov_max[i]  = GrGetMaxValue(slot_index[i], GRIP_CLASS_POV_HAT);*/
         max_value[i][(int)GRIP_CLASS_POV_HAT] = 
               GrGetMaxValue(slot_index[i], GRIP_CLASS_POV_HAT);
         num_hats    = ((int)GrGetMaxIndex(slot_index[i], GRIP_CLASS_POV_HAT) + 1) / 2;
         pov_start = joy[i].num_sticks;
         joy[i].num_sticks += num_hats;
         if (joy[i].num_sticks > MAX_STICKS)  
         {
         	num_hats -= joy[i].num_sticks - MAX_STICKS;
         	joy[i].num_sticks = MAX_STICKS;
         }
         
         /* loop over each pov hat on the device (often just 1) */
         for (stick = pov_start; stick < pov_start + num_hats; stick++)
         {
            stick_class[i][stick] = GRIP_CLASS_POV_HAT;

         	 /* name hat... :( */
            strcpy(tmpstr, name_pov_gen);
            if (num_hats > 1)
            {
               tmpstr[strlen(tmpstr)+2] = '\0';
               tmpstr[strlen(tmpstr)+1] = '1' + stick - pov_start;
               tmpstr[strlen(tmpstr)] = ' ';
            }
            joy[i].stick[stick].name = FIXUP_STR(tmpstr);
            joy[i].stick[stick].flags = JOYFLAG_DIGITAL | JOYFLAG_SIGNED;
            /* name axes */
            joy[i].stick[stick].num_axis = 2;
            GrGetControlName(slot_index[i], GRIP_CLASS_POV_HAT,
                  (GRIP_INDEX)(2*(stick - pov_start)), tmpstr);
            if (tmpstr[0] == '\0')  strcpy(tmpstr, name_x);
            SIMPLIFY_AXIS(tmpstr);
            joy[i].stick[stick].axis[0].name = FIXUP_STR(tmpstr);
            joy[i].stick[stick].axis[0].pos  = 0;
            joy[i].stick[stick].axis[0].d1  = 0;
            joy[i].stick[stick].axis[0].d2  = 0;

            GrGetControlName(slot_index[i], GRIP_CLASS_POV_HAT,
                  (GRIP_INDEX)(2*(stick - pov_start) + 1), tmpstr);
            if (tmpstr[0] == '\0')  strcpy(tmpstr, name_y);
            SIMPLIFY_AXIS(tmpstr);
            joy[i].stick[stick].axis[1].name = FIXUP_STR(tmpstr);
            joy[i].stick[stick].axis[1].pos  = 0;
            joy[i].stick[stick].axis[1].d1  = 0;
            joy[i].stick[stick].axis[1].d2  = 0;
         } /* next hat */
     } /* endif pov hats */


      if (! THROTTLE_AS_BUTTON) { 
      /* Check for Throttles.
         Assume they not grouped, and use 1-axis sticks for them.
       */
      if (ENABLE_THROTTLE &&
          (class_map & (1uL << GRIP_CLASS_THROTTLE)))
      {
         int num_throt = 0;
         int num_throt_sticks = 0;
         int num_throt_extra_axes = 0;
         int stick = 0;
         int throttle_start;
         int throttle_count;

     	 /*throttle_max[i]  = GrGetMaxValue(slot_index[i], GRIP_CLASS_THROTTLE);*/
         max_value[i][(int)GRIP_CLASS_THROTTLE] = 
               GrGetMaxValue(slot_index[i], GRIP_CLASS_THROTTLE);
         num_throt    = (int)GrGetMaxIndex(slot_index[i], GRIP_CLASS_THROTTLE) + 1;

         if (PACK_THROTTLES)
         {
            num_throt_sticks = num_throt / MAX_AXIS;
            num_throt_extra_axes = num_throt % MAX_AXIS;
         } else
         {
            num_throt_sticks = num_throt;
            num_throt_extra_axes = 0;
         }
         
         throttle_start = joy[i].num_sticks;
         joy[i].num_sticks += num_throt_sticks;
         if (joy[i].num_sticks > MAX_STICKS)
         {
         	num_throt_sticks -= joy[i].num_sticks - MAX_STICKS;
         	joy[i].num_sticks = MAX_STICKS;
         }
         /* Account for leftover axes, if packing: */
         if ( num_throt_extra_axes && (joy[i].num_sticks < MAX_STICKS) )
         {
            joy[i].num_sticks += 1;
            num_throt_sticks++;
         } else
         {
         	num_throt_extra_axes = 0;
         }
         
         /* loop over each throttle on the device */
         throttle_count = 0;
         for (stick = throttle_start; stick < throttle_start + num_throt_sticks; stick++)
         {
            int last_axis, a;

            if (PACK_THROTTLES)
            {
               if ( (stick == throttle_start + num_throt_sticks - 1) && 
                    num_throt_extra_axes )
                  last_axis = num_throt_extra_axes;
               else
                  last_axis = MAX_AXIS;
            } else
            {
               last_axis = 1;
            }
            stick_class[i][stick] = GRIP_CLASS_THROTTLE;
            joy[i].stick[stick].flags = JOYFLAG_ANALOG | JOYFLAG_SIGNED;
            joy[i].stick[stick].num_axis = last_axis;
         	 /* name throttle... */
            strcpy(tmpstr, name_throttle_gen);
            if (num_throt_sticks > 1)
            {
               tmpstr[strlen(tmpstr)+2] = '\0';
               tmpstr[strlen(tmpstr)+1] = '1' + stick - throttle_start;
               tmpstr[strlen(tmpstr)] = ' ';
            }
            joy[i].stick[stick].name = FIXUP_STR(tmpstr);

            for (a = 0; a < last_axis; a++)
            {
               /* name axis */
               GrGetControlName(slot_index[i], GRIP_CLASS_THROTTLE, 
                     (GRIP_INDEX)throttle_count, tmpstr);
               if (tmpstr[0] == '\0')
                  strcpy(tmpstr, name_throttle_gen); /* should add number then... */
               joy[i].stick[stick].axis[a].name = FIXUP_STR(tmpstr);

               joy[i].stick[stick].axis[a].pos  = 0;
               joy[i].stick[stick].axis[a].d1  = 0;
               joy[i].stick[stick].axis[a].d2  = 0;
               
               throttle_count++;
            } /* next throttle axis */
         } /* next throttle */
         
         
     } /* endif throttles */

     } /* end if not as button */


      /* DON'T check for velocity - because I don't know what it is, anyway.
         Besides, if the Xterminator doesn't have it, you don't need it.
       */


      /* Be silly and check for buttons... */
      joy[i].num_buttons = 0;
      if (class_map & (1uL << GRIP_CLASS_BUTTON))
      {
      	int b = 0;
      	
         joy[i].num_buttons = (int)GrGetMaxIndex(slot_index[i], GRIP_CLASS_BUTTON) + 1;
         if (joy[i].num_buttons > MAX_BUTTONS)  joy[i].num_buttons = MAX_BUTTONS;

         for (b=0; b < joy[i].num_buttons; b++)
         {
            GrGetControlName(slot_index[i], GRIP_CLASS_BUTTON, (GRIP_INDEX)b, tmpstr);
            if (tmpstr[0] == '\0')  strcpy(tmpstr, name_b[b]);
            joy[i].button[b].name = FIXUP_STR(tmpstr);
            joy[i].button[b].b = 0;
         }
      } /* endif buttons */

      throttle_button_start[i] = joy[i].num_buttons;

      if (THROTTLE_AS_BUTTON) { 
      /* Check for Throttles.
         Assume they not grouped, and use buttons for them.
       */
      if (ENABLE_THROTTLE &&
          (class_map & (1uL << GRIP_CLASS_THROTTLE)))
      {
         int num_throt = 0;
         int throttle = 0;
         int throttle_start;

     	 /*throttle_max[i]  = GrGetMaxValue(slot_index[i], GRIP_CLASS_THROTTLE);*/
         max_value[i][(int)GRIP_CLASS_THROTTLE] = 
               GrGetMaxValue(slot_index[i], GRIP_CLASS_THROTTLE);
         num_throt    = (int)GrGetMaxIndex(slot_index[i], GRIP_CLASS_THROTTLE) + 1;
         throttle_start = joy[i].num_buttons;
         joy[i].num_buttons += num_throt;
         if (joy[i].num_buttons > MAX_BUTTONS)
         {
         	num_throt -= joy[i].num_buttons - MAX_BUTTONS;
         	joy[i].num_buttons = MAX_BUTTONS;
         }
         
         /* loop over each throttle on the device */
         for (throttle = throttle_start; throttle < throttle_start + num_throt; throttle++)
         {
            /* name axis */
            GrGetControlName(slot_index[i], GRIP_CLASS_THROTTLE, 
                  (GRIP_INDEX)(throttle - throttle_start), tmpstr);
            if (tmpstr[0] == '\0')  strcpy(tmpstr, name_throttle_gen);
            joy[i].button[throttle].name = FIXUP_STR(tmpstr);
            joy[i].button[throttle].b = 0;
         } /* next throttle */
     } /* endif throttles */

     } /* end if as button */


   } /* next slot */
   
#if (TEST_CODE)
   if (joy_type == JOY_TYPE_GRIP4)
   {
      if (foo)  fprintf(foo, "joy_type: JOY_TYPE_GRIP4\n");
   } else
   {
      if (foo)  fprintf(foo, "joy_type: JOY_TYPE_GRIP\n");
   }
   if (foo)  fprintf(foo, "UNLOAD_GRIP:%d \t", UNLOAD_GRIP);
   if (foo)  fprintf(foo, "THROTTLE_AS_BUTTON:%d \t", THROTTLE_AS_BUTTON);
   if (foo)  fprintf(foo, "PACK_THROTTLES:%d\n", PACK_THROTTLES);

   if (foo)  fprintf(foo, "ENABLE_AXIS:%d \t", ENABLE_AXIS);
   if (foo)  fprintf(foo, "ENABLE_ANALOG:%d \t", ENABLE_ANALOG);
   if (foo)  fprintf(foo, "ENABLE_POV:%d \t", ENABLE_POV);
   if (foo)  fprintf(foo, "ENABLE_THROTTLE:%d\n", ENABLE_THROTTLE);

   if (foo)  fprintf(foo, "MAX_JOYSTICKS:%d \t", MAX_JOYSTICKS);
   if (foo)  fprintf(foo, "MAX_AXIS:%d \t", MAX_AXIS);
   if (foo)  fprintf(foo, "MAX_BUTTONS:%d \t", MAX_BUTTONS);
   if (foo)  fprintf(foo, "MAX_STICKS:%d\n", MAX_STICKS);

   if (foo)  fprintf(foo, "\n********************\n");
   if (foo)  fprint_grip(foo);
   if (foo)  fprintf(foo, "\n********************\n");
   if (foo)  fprint_joy(foo);
   if (foo)  (void)fclose(foo);
#endif /*TEST_CODE*/

   return 0;
}



/* grip_exit:
 *  Shuts down the driver.
 */
static void grip_exit()
{
   if (!UNLOAD_GRIP)  return;

   free_text();   
   GrShutdown();
   (void)grip_unlink();
}



/* grip_poll:
 *  Updates the joystick status variables.
 *  Assumes GrIP driver has been validly initialized.
 */
static int grip_poll()
{
   int i;
   int all_good;
   
   all_good = 0;
   (void)GrRefresh(0);
   
   for (i=0; i<num_joysticks; i++) {
      if (read_grip(i) == 1)
	     all_good = -1;
   }

   return all_good;
}


static int already_linked = 0;

/* The following intiallization routine is mostly taken from
 * the Gravis SDK. Note that the file GRIP.GLL is
 * required in either the path, the current directory,
 * or C:\GRIP .
 */
static int grip_link()
{
  char lib[128];              // Full path for GrIP library
  FILE* flib;                 // File handle for library
  long size;                  // Size of library file (bytes)
  void* image;                // Memory image of library file

  // Search for the GrIP library: GRIP.GLL
  //
  // First search in the current directory, then search in any directories
  // specified by the "SET GRIP=" environment variable, then search in any
  // directories in the path, and finally search in "C:\GRIP\".
  //
  // Note that the filename for the library ("GRIP.GLL"), the environment
  // variable name ("GRIP") and the default install directory ("C:\GRIP\")
  // are defined in the module GRIP.C as GrLibName, GrLibEnv, and
  // GrLibDir respectively.

  if (already_linked)  return 0;
  
  lib[0] = '\0';
  strcpy(lib, searchpath(GrLibName));

  if (lib[0] == '\0') {
     strcpy(lib, getenv(GrLibEnv));
     if (lib[strlen(lib) - 1] != '\\') strcat(lib, "\\");
     strcat(lib, GrLibName);
  }

  if (!__file_exists(lib)) {
     strcpy(lib, GrLibDir);
     strcat(lib, GrLibName);
  }

  // Open the loadable library file and determine its size
  if ( (flib = fopen(lib, "rb")) == 0 )
    {
       printf("\r\nERROR: Unable to load library file \"%s\"\r\n", lib);
       return -1;
    }
  (void)fseek(flib, 0, SEEK_END);
  size = ftell(flib);

  // Allocate memory for the library and read it all into memory
  //
  // (If the size is greater than 32k, then this is definitely not
  // the correct file, so signal an error.)
  if (size > 32*1024L)
    {
       printf("\r\nGRIP.GLL ERROR: Invalid library file \"%s\" (File too large)\r\n", lib);
       return -1;
    }
  image = (void*)malloc((size_t)size);
  if (image == 0)
    {
       printf("\r\nERROR: Unable to allocate sufficient memory for library \"%s\"\r\n", lib);
       return -1;
    }
  (void)fseek(flib, 0, SEEK_SET);
  (void)fread(image, 1, (size_t)size, flib);
  (void)fclose(flib);

  // Link to the library (and free up the temporary memory)
  if (!GrLink(image, (GRIP_VALUE)size))
    {
       printf("\r\nGRIP.GLL ERROR: Unable to initialize loadable library!\r\n");
       free(image);
       return -1;
    }
  free(image);

  already_linked = 1;
  return 0;
}	



static int grip_unlink()
{
	if (!UNLOAD_GRIP)  return 0;

    already_linked = 0;
	if (!GrUnlink())  return -1;
	return 0;
}



/***** Text Heap functions *****/
/* This could be a full-featured data abstraction, but I really
   just threw it together to hack a solution to the inexplicable
   bugs I was getting.
*/

/* Return a newly-allocated, unshared copy of the argument.
   Assumes that txt is a valid string, if not null. 
 */
static char * 
new_text(const char * const txt)
{
	text_heap *next;
	
	if (!txt)  return NULL;

	next = (text_heap *)malloc(sizeof(text_heap));
	if (!next)
	{
		/* Better to return NULL, or a shared "" ?*/
		return "";
		/*return NULL*/
	}
	next->data = (char *)malloc(sizeof(char)*(strlen(txt) + 1));
	if (!next->data)
	{
		free(next);
		/* Better to return NULL, or a shared "" ?*/
		return "";
		/*return NULL*/
	}
	
	next->last = texth;
	texth = next;
	strcpy(texth->data, txt);
	return texth->data;
}

static int
text_heap_size(const text_heap * const th)
{
	if (!th)  return 0;
	return 1 + text_heap_size(th->last);
}

/* Free all text_heap elements in global text_root */
static void   
free_text(void)
{
	text_heap *next;

	while (texth)
	{
		next = texth->last;
		free(texth);
		texth = next;
	}
}



/***** Debug output routines *****/
#if TEST_CODE

static void
fprint_grip(FILE * const fp)
{
	int i;
	GRIP_BITFIELD slotmap;
	
	if (!fp)  return;
	fprintf(fp, "GrIP Info:\n");
	slotmap = GrGetSlotMap();
	fprintf(fp, "\tSlot map : %s\n", binstr((unsigned int)slotmap));
	for (i = 1; i < 31; i++)
	{
		if (slotmap & (1uL << i))
		{
			fprintf(fp, "\t\tSlot %d:\n", i);
			fprint_slot(fp, (GRIP_SLOT)i);
		}
	}
}

static char *grip_classname[] = {"GRIP_CLASS_RESERVED", "GRIP_CLASS_BUTTON", "GRIP_CLASS_AXIS",
      "GRIP_CLASS_POV_HAT", "GRIP_CLASS_VELOCITY", "GRIP_CLASS_THROTTLE",
      "GRIP_CLASS_ANALOG"};
static void
fprint_slot(FILE * const fp, const GRIP_SLOT slot)
{
	GRIP_BITFIELD classmap;
	int i;
	char tmpstr[512] = "";
			
	classmap = GrGetClassMap(slot);
	

	i = (int)GrGetProductName(slot, tmpstr);
	fprintf(fp, "\t\t\tName     : '%s'\n", tmpstr);
	fprintf(fp, "\t\t\tVersion  : %d\n", i);
	GrGetVendorName(slot, tmpstr);
	fprintf(fp, "\t\t\tVendor   : '%s'\n", tmpstr);
	fprintf(fp, "\t\t\tClass map: %s\n", binstr((unsigned int)classmap));
	fprintf(fp, "\t\t\tOEM   map: %s\n", binstr((unsigned int)GrGetOEMClassMap(slot)));

	for (i = GRIP_CLASS_BUTTON; i <= (int)GRIP_CLASS_ANALOG; i++)
	{
		if (!(classmap & (1uL << i)))
		{
			fprintf(fp, "\t\t\t%s (0)\n", grip_classname[i]);
			continue;
		}
		fprint_class(fp, slot, (GRIP_CLASS)i);
	}
}

static void
fprint_class(FILE * const fp, const GRIP_SLOT slot, const GRIP_CLASS gclass)
{
	int num_controls, i;
	char tmpstr[512] = "";
	
	num_controls = (int)GrGetMaxIndex(slot, gclass) + 1;
	fprintf(fp, "\t\t\t%s (%d):\n", grip_classname[(int)gclass], num_controls);
	fprintf(fp, "\t\t\t\tMax value: %d\n", (int)GrGetMaxValue(slot, gclass));
	for (i = 0; i < num_controls; i++)
	{
		GrGetControlName(slot, gclass, (GRIP_INDEX)i, tmpstr);
		fprintf(fp, "\t\t\t\t(%X) '%s': %d\n", 
				(unsigned int)i,
				tmpstr,
				(int)GrGetValue(slot, gclass, (GRIP_INDEX)i));
	}
}


static char *
binstr(unsigned int bitfield)
{
	int i, bit;
	static char s[37];
	unsigned int b;
	
	s[36] = '\0';
	s[35] = '\0';
	bit = 31;
	for (i = 34; i >= 0; i--)
	{
		if (!(bit % 8) && (bit != 0) && (i != 0))
		{
			s[i] = ' ';
			i--;
		}
		if (bitfield % 2)
		{
			s[i] = '1';
		} else
		{
			s[i] = '0';
		}
		bitfield >>= 1;
		bit--;
	}
	return s;
}

static char *
binstr_long(unsigned long bitfield)
{
	int i, bit;
	static char s[73];
	
	s[72] = '\0';
	s[71] = '\0';
	bit = 63;
	for (i = 70; i >= 0; i--)
	{
		if (!(bit % 8) && (bit != 0) && (i != 0))
		{
			s[i] = ' ';
			i--;
		}
		if (bitfield % 2)
		{
			s[i] = '1';
		} else
		{
			s[i] = '0';
		}
		bitfield >>= 1;
		bit --;
	}
	return s;
}


static void
fprint_joy(FILE * const fp)
{
	int i;

	if (!fp)  return;

	fprintf(fp, "Allegro reports %d joysticks.\n", num_joysticks);
	for (i = 0; i < num_joysticks; i++)
	{
		fprintf(fp, "\nJoystick %d:\n", i);
		fprint_joystick(fp, joy[i]);
	}
}

static void
fprint_joystick(FILE * const fp, const JOYSTICK_INFO j)
{
	int i;
	
	fprintf(fp, "\tFlags: ");
	if (j.flags & JOYFLAG_DIGITAL)  fprintf(fp, "JOYFLAG_DIGITAL ");
	if (j.flags & JOYFLAG_ANALOG)  fprintf(fp, "JOYFLAG_ANALOG ");
	if (j.flags & JOYFLAG_CALIB_DIGITAL)  fprintf(fp, "JOYFLAG_CALIB_DIGITAL ");
	if (j.flags & JOYFLAG_ANALOG)  fprintf(fp, "JOYFLAG_CALIB_ANALOG ");
	if (j.flags & JOYFLAG_CALIBRATE)  fprintf(fp, "JOYFLAG_CALIBRATE ");
	if (j.flags & JOYFLAG_SIGNED)  fprintf(fp, "JOYFLAG_SIGNED ");
	if (j.flags & JOYFLAG_UNSIGNED)  fprintf(fp, "JOYFLAG_UNSIGNED ");
	fprintf(fp,"\n");
	/*if (j.num_buttons > 0)*/  fprintf(fp, "\tButtons (%d):\n", j.num_buttons);
	for (i = 0; i < j.num_buttons; i++)
	{
		fprint_button(fp, j.button[i]);
	}
	/*if (j.num_sticks > 0)*/  fprintf(fp, "\tSticks (%d):\n", j.num_sticks);
	for (i = 0; i < j.num_sticks; i++)
	{
		fprint_stick(fp, j.stick[i]);
	}
}

static void
fprint_button(FILE * const fp, const JOYSTICK_BUTTON_INFO button)
{
	fprintf(fp, "\t\t'%s' (%d)\n", button.name, (int)button.b);
}

static void
fprint_stick(FILE * const fp, const JOYSTICK_STICK_INFO s)
{
	int i;

	fprintf(fp, "\t\t'%s':\n", s.name);
	
	fprintf(fp, "\t\t\tFlags: ");
	if (s.flags & JOYFLAG_DIGITAL)  fprintf(fp, "JOYFLAG_DIGITAL ");
	if (s.flags & JOYFLAG_ANALOG)  fprintf(fp, "JOYFLAG_ANALOG ");
	if (s.flags & JOYFLAG_CALIB_DIGITAL)  fprintf(fp, "JOYFLAG_CALIB_DIGITAL ");
	if (s.flags & JOYFLAG_ANALOG)  fprintf(fp, "JOYFLAG_CALIB_ANALOG ");
	if (s.flags & JOYFLAG_CALIBRATE)  fprintf(fp, "JOYFLAG_CALIBRATE ");
	if (s.flags & JOYFLAG_SIGNED)  fprintf(fp, "JOYFLAG_SIGNED ");
	if (s.flags & JOYFLAG_UNSIGNED)  fprintf(fp, "JOYFLAG_UNSIGNED ");
	fprintf(fp,"\n");

	fprintf(fp, "\t\t\tAxes (%d):\n", (int)s.num_axis);
	for (i = 0; i < s.num_axis; i++)
	{
		fprint_axis(fp, s.axis[i]);
	}
}

static void
fprint_axis(FILE * const fp, const JOYSTICK_AXIS_INFO a)
{
	fprintf(fp, "\t\t\t\t'%s': %d (d1 = %d, d2 = %d)\n", 
	   a.name, (int)a.pos, (int)a.d1, (int)a.d2);
}

#endif /*TEST_CODE*/

